'use strict';

var ParserHelpers = require('webpack/lib/ParserHelpers');
var AsyncRequireDependenciesBlock = require('./AsyncRequireDependenciesBlock');
var OpenUI5LazyInstanceDependency = require('./OpenUI5LazyInstanceDependency');
var OpenUI5ViewDependency = require('./OpenUI5ViewDependency');
var OpenUI5RequireDependency = require('./RequireDependency');
var OpenUI5RequireItemDependency = require('./RequireItemDependency');
var OpenUI5RequireContextDependency = require('./OpenUI5RequireContextDependency');
var ContextDependencyHelpers = require('./ContextDependencyHelpers');

class OpenUI5RequireDependencyParserPlugin {
  constructor(options) {
    this.options = options;
  }

  newRequireDependency(range, moduleRange, arrayRange, functionRange) {
    return new OpenUI5RequireDependency(range, moduleRange, arrayRange, functionRange);
  }

  apply(parser) {
    var _this = this;

    var options = this.options || {};
    var globals = {
      'sap.ui.core.theming.Parameters': 'sap/ui/core/theming/Parameters',
      'sap.ui.core.LabelEnablement': 'sap/ui/core/LabelEnablement',
      'sap.ui.core.HTML': 'sap/ui/core/HTML',
      'sap.m.CheckBox': 'sap/m/CheckBox',
      'sap.viz.ui5.theming.Util': 'sap/viz/ui5/theming/Util',
      'sap.ui.unified.DateRange': 'sap/ui/unified/DateRange',
      'sap.ui.core.IconPool': 'sap/ui/core/IconPool',
      'sap.m.IconTabHeader': 'sap/m/IconTabHeader'
    };

    var processItem = function processItem(expr, param, range) {
      if (param.isString()) {
        var item = param.string;
        if (item.substr(0, 10) !== 'jquery.sap' && item.substr(0, 2) !== './') {
          item = item.replace(/\./g, '/');
        }
        var dep = new OpenUI5RequireItemDependency(item, range);
        dep.loc = expr.loc;
        dep.optional = !!parser.scope.inTry;
        parser.state.current.addDependency(dep);
        return true;
      }
      return false;
    };

    var processContext = function processContext(expr, param) {
      var dep = ContextDependencyHelpers.create(OpenUI5RequireContextDependency, expr.range, param, expr, options);
      if (!dep) return false;
      dep.loc = expr.loc;
      dep.optional = !!parser.scope.inTry;
      parser.state.current.addDependency(dep);
      return true;
    };

    var processArray = function processArray(expr, param) {
      param.items.forEach(function (param) {
        processItem(expr, param, param.range);
      });
      return true;
    };

    var processGlobal = function processGlobal(expr, request) {
      var dep = new OpenUI5RequireItemDependency(request, expr.range, true);
      dep.loc = expr.loc;
      dep.optional = !!parser.scope.inTry;
      parser.state.current.addDependency(dep);
      return true;
    };

    var processRemove = ParserHelpers.toConstantDependency(parser, process.env.NODE_ENV === 'production' ? '' : 'console.warn("UI5 tried to dynamically require a module. If it doesn\'t exist in the bundle a error might follow.")');

    parser.hooks.call.for('sap.ui.require').tap('OpenUI5Plugin', function (expr) {
      var module = void 0;
      var array = void 0;
      var fn = void 0;

      if (expr.arguments[0].type === 'Literal') {
        module = expr.arguments[0];
      } else if (expr.arguments[0].type === 'ArrayExpression') {
        array = expr.arguments[0];
      } else {
        // console.warn(`OpenUI5 webpack plugin: cannot handle dynamic dependency (type ${expr.arguments[0].type}) for sap.ui.require. Skipping dependency resolution.`);
        return false;
      }

      if (expr.arguments[1]) {
        fn = expr.arguments[1];
      }

      var param = parser.evaluateExpression(expr.arguments[0]);
      if (module) {
        var _result = processItem(expr, param, expr.range);
        return _result;
      }

      var result = processArray(expr, param);
      if (!result) return false;

      // If a callback function is given, the modules should be loaded asynchronously on request only
      if (fn && fn.type === 'FunctionExpression') {
        if (fn.body.type === 'BlockStatement') {
          parser.walkStatement(fn.body);
        } else {
          parser.walkExpression(fn.body);
        }

        var _dep = new AsyncRequireDependenciesBlock(expr, expr.arguments[0].range, expr.arguments[1].range, parser.state.module, expr.loc);
        parser.state.current.addBlock(_dep);
        return true;
      }

      var dep = _this.newRequireDependency(expr.range, module ? module.range : null, array ? array.range : null, fn ? fn.range : null);
      dep.loc = expr.loc;
      parser.state.current.addDependency(dep);
      return true;
    });

    parser.hooks.call.for('sap.ui.requireSync').tap('OpenUI5Plugin', function (expr) {
      var param = parser.evaluateExpression(expr.arguments[0]);

      var result = processItem(expr, param, expr.range);
      if (!result) {
        result = processContext(expr, param);
      }
      return result;
    });

    parser.hooks.call.for('lazyInstanceof').tap('OpenUI5Plugin', function (expr) {
      var param1 = parser.evaluateExpression(expr.arguments[1]);
      if (param1.string !== 'sap/ui/app/Application') {
        var dep = new OpenUI5LazyInstanceDependency(param1.string, expr.arguments[0].name, expr.range);
        dep.loc = expr.loc;
        dep.optional = !!parser.scope.inTry;
        parser.state.current.addDependency(dep);
        return true;
      }
      return false;
    });

    parser.hooks.call.for('createView').tap('OpenUI5Plugin', function (expr) {
      var dep = void 0;
      var param0 = parser.evaluateExpression(expr.arguments[0]);
      if (expr.arguments[0].type === 'Literal') {
        dep = new OpenUI5ViewDependency(param0.string, expr.arguments[1].name, expr.range);
        if (!dep) {
          return false;
        }
      } else if (expr.arguments[0].type === 'Identifier') {
        dep = ContextDependencyHelpers.createView(OpenUI5RequireContextDependency, expr.range, param0, expr, options);
      } else {
        return false;
      }
      dep.loc = expr.loc;
      dep.optional = !!parser.scope.inTry;
      parser.state.current.addDependency(dep);
      return true;
    });

    parser.hooks.call.for('jQuery.sap.require').tap('OpenUI5Plugin', function (expr) {
      var param = parser.evaluateExpression(expr.arguments[0]);
      if (expr.arguments[0].type === 'Literal') {
        processItem(expr, param, expr.range);
        return true;
      }
      processRemove(expr, param);
      return false;
    });

    parser.hooks.call.for('$.sap.require').tap('OpenUI5Plugin', function (expr) {
      var param = parser.evaluateExpression(expr.arguments[0]);
      if (expr.arguments[0].type === 'Literal') {
        processItem(expr, param, expr.range);
        return true;
      }
      return false;
    });

    Object.keys(globals).forEach(function (key) {
      parser.hooks.expression.for(key).tap('OpenUI5Plugin', function (expr) {
        return processGlobal(expr, globals[key]);
      });
    });
  }
}
module.exports = OpenUI5RequireDependencyParserPlugin;