'use strict';

var RequireDependenciesBlock = require('./RequireDependenciesBlock');
var RequireItemDependency = require('./RequireItemDependency');
var ContextDependencyHelpers = require('./ContextDependencyHelpers');

class RequireDependencyParserPlugin {
  constructor(options) {
    this.options = options;
  }

  apply(parser) {
    var options = this.options || {};

    var processItem = function processItem(expr, param, range) {
      if (param.isString()) {
        var item = param.string;
        if (!item.match(/j[Qq]uery\.sap/) && item.substr(0, 2) !== './') {
          item = item.replace(/\./g, '/');
        }
        var dep = new RequireItemDependency(item, range);
        dep.loc = expr.loc;
        dep.optional = !!parser.scope.inTry;
        parser.state.current.addDependency(dep);
        return true;
      }
      return false;
    };

    var processContext = function processContext(expr, param) {
      var dep = ContextDependencyHelpers.create(expr.range, param, expr, options);
      if (!dep) return false;
      dep.loc = expr.loc;
      dep.optional = !!parser.scope.inTry;
      parser.state.current.addDependency(dep);
      return true;
    };

    var processArray = function processArray(expr, param) {
      param.items.forEach(function (param) {
        processItem(expr, param, param.range);
      });
      return true;
    };

    parser.hooks.call.for('sap.ui.require').tap('OpenUI5Plugin', function (expr) {
      if (expr.arguments[0].type === 'Literal') {
        // Synchronous retrieval of module only one string argument
        var _module = expr.arguments[0];
        var param = parser.evaluateExpression(_module);
        return processItem(expr, param, expr.range);
      } else if (expr.arguments[0].type === 'ArrayExpression') {
        // Async loading of module and optionally execution of callback function
        var array = expr.arguments[0];
        var _param = parser.evaluateExpression(array);
        var fn = expr.arguments[1];

        var dependenciesItems = _param ? _param.items : [_param];
        var dep = new RequireDependenciesBlock(expr, array.range, fn ? fn.range : null, parser.state.module, expr.loc, dependenciesItems[0].string);
        var old = parser.state.current;
        parser.state.current = dep;

        try {
          var result = processArray(expr, _param);
          if (!result) {
            return;
          }

          if (fn && fn.type === 'FunctionExpression') {
            if (fn.body.type === 'BlockStatement') {
              parser.walkStatement(fn.body);
            } else {
              parser.walkExpression(fn.body);
            }
          }
          old.addBlock(dep);
        } finally {
          parser.state.current = old;
        }

        return true;
      }
    });

    parser.hooks.call.for('sap.ui.requireSync').tap('OpenUI5Plugin', function (expr) {
      var param = parser.evaluateExpression(expr.arguments[0]);

      var result = processItem(expr, param, expr.range);
      if (!result) {
        result = processContext(expr, param);
      }
      return result;
    });
  }
}
module.exports = RequireDependencyParserPlugin;